library lemon_dispatcher;

class Dispatcher {
  final List<Handler> _handlers = [];

  void dispatch<T>([var value]) {
    int handlersFound = 0;

    for (int i = 0; i < _handlers.length; i++) {
      if (!_handlers[i].enabled) continue;

      if (value != null && !_handlers[i].canHandle(value)) continue;

      if (value == null && !_handlers[i].canHandleType<T>()) continue;

      handlersFound++;

      Future futureResult = _handlers[i].handle(value).catchError((error) {
        dispatch(error);
        return null;
      });

      futureResult.then((output) {
        if (output == null) return;

        if (_handlers[i].canHandle(output)) {
          _handlers[i].enabled = false;
          print("handler disabled because it returns own handle type");
          return;
        }

        dispatch(output);
      });
    }

    if (handlersFound == 0) {
      print("Warning no handlers found for $value");
    }
  }

  void register<T>(HandlerFunction<T> _handlerFunction) {
    _handlers.add(Handler<T>(_handlerFunction));
  }
}

class Handler<T> {
  bool enabled = true;
  HandlerFunction<T> _handlerFunction;

  Handler(this._handlerFunction);

  Future handle(var value) async {
    return await _handlerFunction(value);
  }

  bool canHandle(var value) {
    return value is T;
  }

  bool canHandleType<J>() {
    return J == T;
  }
}

typedef Future<dynamic> HandlerFunction<E>(E element);
